import pytest

def pytest_addoption(parser):
    """
    Add command-line options to specify MPI and coverage configuration
    """
    parser.addoption("--no-build", action="store_true", default=False,
                    help="do not build the project (use system installed version)")

    parser.addoption("--build-only", action="store_true", default=False,
                    help="just build, do not run any tests")

    parser.addoption("--shell", action="store_true", default=False,
                    help="start a shell with the installed package properly set up in the path")
                    
    parser.addoption("--show-build-log", action="store_true",
                    help="show build output rather than using a log file")
    
    parser.addoption("--with-coverage", action="store_true", default=False,
                    help="report coverage of project code to a .coverage file")
                          
    parser.addoption("--html-cov", action="store_true", default=False,
                    help="write html coverage reports to build/coverage")
                          
    parser.addoption('--cov-config', action='store', default='.coveragerc',
                    metavar='path',
                    help=('config file for coverage, default: .coveragerc; '
                          'see http://coverage.readthedocs.io/en/coverage-4.3.4/config.html'))
    
def pytest_collection_modifyitems(session, config, items):
    """
    Modify the ordering of tests, such that the ordering will be 
    well-defined across all ranks running
    """
    items[:] = sorted(items, key=lambda x: str(x))



import pytest

def pytest_addoption(parser):
    """
    Add command-line options to specify MPI and coverage configuration
    """
    parser.addoption("--mpirun", default=None, nargs='?', const="mpirun -n 4",
                    help="launcher for MPI, e.g. mpirun -n 4")

    parser.addoption("--mpisub", action="store_true", default=False,
                    help="run process as a mpisub")
        
    parser.addoption("--mpisub-site-dir", default=None, help="site-dir in mpisub")
    



from .. import conftest
from ..tester import Tester as BaseTester

import pytest
from mpi4py import MPI
import traceback
import warnings
import sys
import os
import contextlib
import time
from argparse import ArgumentParser

if sys.version_info[0] == 2:
    from cStringIO import StringIO
else:
    from io import StringIO
import re

def fix_titles(s):
    pattern = '=====+'
    return re.sub(pattern, lambda x: x.group(0).replace('=', '-'), s)

class Rotator(object):
    """ in a rotator every range runs in terms """
    def __init__(self, comm):
        self.comm = comm
    def __enter__(self):
        self.comm.Barrier()
        for i in range(self.comm.rank):
            self.comm.Barrier()
    def __exit__(self, type, value, tb):
        for i in range(self.comm.rank, self.comm.size):
            self.comm.Barrier()
        self.comm.Barrier()

def MPITest(commsize):
    """
    A decorator that repeatedly calls the wrapped function,
    with communicators of varying sizes.

    This converts the test to a generator test; therefore the
    underlyig test shall not be a generator test.
    
    Parameters
    ----------
    commsize: scalar or tuple
        Sizes of communicator to use

    Usage
    -----
    @MPITest(commsize=[1, 2, 3])
    def test_stuff(comm):
        pass
    """
    if not isinstance(commsize, (tuple, list)):
        commsize = (commsize,)

    sizes = sorted(list(commsize))

    def dec(func):
        
        @pytest.mark.parametrize("size", sizes)
        def wrapped(size, *args):
            if MPI.COMM_WORLD.size < size: 
                pytest.skip("Test skipped because world is too small. Include the test with mpirun -n %d" % (size))
                
            color = 0 if MPI.COMM_WORLD.rank < size else 1
            comm = MPI.COMM_WORLD.Split(color)
            try:
                if color == 0:
                    rt = func(*args, comm=comm)
                if color == 1:
                    rt = None
                    #pytest.skip("rank %d not needed for comm of size %d" %(MPI.COMM_WORLD.rank, size))
            finally:
                MPI.COMM_WORLD.barrier()
                
            return rt
        wrapped.__name__ = func.__name__
        return wrapped
    return dec
    
    
def MPIWorld(NTask, required=1, optional=False):
    """ 
    A decorator that repeatedly calls the wrapped function,
    with communicators of varying sizes.
    
    .. note:: Deprecated 
        See :func:`MPITest` instead

    Parameters
    ----------
    NTask : scalar or tuple
        Size of communicators ot use
    required : scalar or tuple
        Required sizes of communicators. If the MPI_WORLD is insufficient, a Error is raised
        to abort the tests.
    optional : boolean
        If requirement not satistied, skip the test.
    """
    warnings.warn("This function is deprecated, use MPITest instead.", DeprecationWarning)

    if not isinstance(NTask, (tuple, list)):
        NTask = (NTask,)

    if not isinstance(required, (tuple, list)):
        required = (required,)

    maxsize = max(required)
    if MPI.COMM_WORLD.size < maxsize and not optional:
        raise ValueError("Test Failed because the world is too small. Increase to mpirun -n %d, current size = %d" % (maxsize, MPI.COMM_WORLD.size))
        
    sizes = sorted(set(list(required) + list(NTask)))
    def dec(func):
        
        @pytest.mark.parametrize("size", sizes)
        def wrapped(size, *args):
            if MPI.COMM_WORLD.size < size: 
                pytest.skip("Test skipped because world is too small. Include the test with mpirun -n %d" % (size))

            color = 0 if MPI.COMM_WORLD.rank < size else 1
            comm = MPI.COMM_WORLD.Split(color)
            
            if color == 0:
                rt = func(*args, comm=comm)

            MPI.COMM_WORLD.barrier()
            if color == 1:
                #pytest.skip("rank %d not needed for comm of size %d" %(MPI.COMM_WORLD.rank, size))
                rt = None
            return rt
        wrapped.__name__ = func.__name__
        return wrapped
    return dec

class Tester(BaseTester):
    """        
    Run MPI-enabled tests using pytest, building the project first.
    
    Examples::
        $ python runtests.py my/module --mpirun
        $ python runtests.py my/module/tests/test_abc.py --mpirun
        $ python runtests.py my/module --mpirun="mpirun -np 4"
        $ python runtests.py --mpirun="mpirun -np 4"
    """
    plugins = [conftest.build, conftest.mpi]
    
    def __init__(self, *args, **kwargs):
        """
        Parameters
        ----------
        package_file : str
            the path to the main directory of the source package
        module : str
            the name of the package to test
        extra path : list of str
            extra paths to include on PATH when building
        """
        super(Tester, self).__init__(*args, **kwargs)
        self.comm = MPI.COMM_WORLD
        
    def main(self, argv):
        
        argv.append('-x')
        config = self._get_pytest_config(argv)
        args = config.known_args_namespace
        
        # print help and exit
        if args.help:
            return config.hook.pytest_cmdline_main(config=config)
        
        # import project from system path
        args.pyargs = True
                        
        # mpi subs do not build!
        if args.mpisub:
            args.no_build = True # master does the building
                        
        # build the project, returning the site directory
        site_dir = self._do_build(args)

        if args.shell:
            self._do_shell(args, config)

        if args.mpirun:
            
            # extract the mpirun run argument
            parser = ArgumentParser(add_help=False)
            parser.add_argument("--mpirun", default=None, const='mpirun -n 4', nargs='?')
            args, additional = parser.parse_known_args()

            # make the test directory exists
            self._initialize_testdir()

            # now call with mpirun
            mpirun = args.mpirun.split()
            cmdargs = [sys.executable, sys.argv[0], '--mpisub', '--mpisub-site-dir=' + site_dir]
            os.execvp(mpirun[0], mpirun + cmdargs + additional)
            sys.exit(1)

        if args.build_only:
            sys.exit(0)
                    
        # fix the path of the modules we are testing
        config.args = self._fix_test_paths(site_dir, config.args) 

        # reset the output
        if args.mpisub:

            self.oldstdout = sys.stdout
            self.oldstderr = sys.stderr
            newstdout = StringIO()
            newstderr = StringIO()

            if self.comm.rank != 0:
                sys.stdout = newstdout
                sys.stderr = newstderr

        # test kwargs
        kws = {}
        kws['with_coverage'] = args.with_coverage
        kws['config_file'] = args.cov_config
        kws['html_cov'] = args.html_cov
        kws['comm'] = self.comm

        # run the tests
        code = None
        with self._run_from_testdir(args):
            code = self._test(config, **kws)
            
        self.comm.barrier()
        
        if args.mpisub:
            if code != 0:
                # if any rank has a failure, print the error and abort the world.
                self._sleep()
                self.oldstderr.write("Test Failure due to rank %d\n" % self.comm.rank)
                self.oldstderr.write(newstdout.getvalue())
                self.oldstderr.write(newstderr.getvalue())
                self.oldstderr.flush()
                self.comm.Abort(-1)

            self.comm.barrier()
            with Rotator(self.comm):
                if self.comm.rank != 0:
                    self.oldstderr.write("\n")
                    self.oldstderr.write("=" * 32 + " Rank %d / %d " % (self.comm.rank, self.comm.size) + "=" * 32)
                    self.oldstderr.write("\n")
                    self.oldstderr.write(fix_titles(newstdout.getvalue()))
                    self.oldstderr.write(fix_titles(newstderr.getvalue()))
                    self.oldstderr.flush()

            sys.exit(0)

        else:
            sys.exit(code)
        
    def _do_build(self, args):
        
        site_dir = super(Tester, self)._do_build(args)
            
        if args.mpisub_site_dir:
            site_dir = args.mpisub_site_dir
            sys.path.insert(0, site_dir)
            os.environ['PYTHONPATH'] = site_dir
        return site_dir
        
    def _sleep(self):
        time.sleep(0.04 * self.comm.rank)
        
    @contextlib.contextmanager
    def _run_from_testdir(self, args):
        
        cwd = os.getcwd()
        self._initialize_testdir()
        
        try:
            if args.mpisub:
                assert(os.path.exists(self.TEST_DIR))
                self.comm.barrier()
            os.chdir(self.TEST_DIR)
            yield
        except:
            if args.mpisub:
                self._sleep()
                self.oldstderr.write("Fatal Error on Rank %d\n" % self.comm.rank)
                self.oldstderr.write(traceback.format_exc())
                self.oldstderr.flush()
                self.comm.Abort(-1)
            else:
                traceback.print_exc()
                sys.exit(1)
        finally:
            os.chdir(cwd)
    



from .coverage import Coverage
from . import conftest

import pytest
import traceback
import sys
import os
import contextlib
import shutil
import subprocess
import time


class Tester(object):
    """
    Run tests using pytest, building a fresh version of the project first.
    
    Examples::
        $ python runtests.py my/module 
        $ python runtests.py my/module/tests/test_abc.py
        $ python runtests.py
    """
    plugins = [conftest.build]
    
    def __init__(self, package_file, module,
            extra_path =['/usr/lib/ccache', '/usr/lib/f90cache',
                         '/usr/local/lib/ccache', '/usr/local/lib/f90cache']
        ):
        """
        Parameters
        ----------
        package_file : str
            the path to the main directory of the source package
        module : str
            the name of the package to test
        extra path : list of str
            extra paths to include on PATH when building
        """
        self.ROOT_DIR = os.path.abspath(os.path.join(os.path.dirname(package_file)))
        self.PROJECT_ROOT_FILES = [package_file, 'setup.py', 'LICENSE']
        self.EXTRA_PATH = extra_path
        self.PROJECT_MODULE = module
        self.TEST_DIR = os.path.join(self.ROOT_DIR, 'build', 'test')
        
    def main(self, argv):
        """
        The main function to run the tests
        
        Parameters
        ----------
        argv : list of str
            the command-line arguments -- should be equal to ``sys.argv[1:]``
        """
        # initialize the pytest configuration from the command-line args
        config = self._get_pytest_config(argv)
        args = config.known_args_namespace
        
        # print help and exit
        if args.help:
            return config.hook.pytest_cmdline_main(config=config)
        
        # import project from system path
        # this forces pytest to import the freshly built package
        args.pyargs = True
                        
        # build the project, returning the site directory
        site_dir = self._do_build(args)

        if args.shell:
            self._do_shell(args, config)

        if args.build_only:
            sys.exit(0)
                    
        # fix the path of the modules we are testing
        # so they point to site_dir
        config.args = self._fix_test_paths(site_dir, config.args) 

        # make sure the test directory exists
        self._initialize_testdir()

        # extract the coverage-related options
        covargs = {}
        covargs['with_coverage'] = args.with_coverage
        covargs['config_file'] = args.cov_config
        covargs['html_cov'] = args.html_cov

        # run the tests
        code = None
        with self._run_from_testdir(args):
            code = self._test(config, **covargs)
        
        # and exit  
        sys.exit(code)
                
    def _test(self, config, **kwargs):
        """
        Run the actual tests with optional coverage -- a wrapper around 
        the pytest calling sequence
        
        Parameters
        ----------
        config : 
            the pytest configuration object
        kwargs : 
            additional keywords to pass to the Coverage class
        """
        try:
            with Coverage(self.PROJECT_MODULE, root=self.ROOT_DIR, **kwargs):
                config.pluginmanager.check_pending()
                return config.hook.pytest_cmdline_main(config=config)
        finally:
            config._ensure_unconfigure()
            
    def _do_build(self, args):
        """
        Build the project and return the site directory in the 
        build/ directory
        """
        site_dir='.' # use source directory by default
        if not args.no_build:
            site_dir = self._build_project(args)
            sys.path.insert(0, site_dir)
            os.environ['PYTHONPATH'] = site_dir
        
        return site_dir

    def _do_shell(self, args, config):
        capman = config.pluginmanager.getplugin('capturemanager')
        if capman:
            out, err = capman.suspendcapture(in_=True)
        shell = os.environ.get('SHELL', 'sh')
        print("Spawning a Unix shell...")
        with self._run_from_testdir(args):
            os.execv(shell, [shell] + config.args[1:])
            sys.exit(1)

    @contextlib.contextmanager
    def _run_from_testdir(self, args):
        """
        Context manager to safely change directory
        to ``build/test``
        """
        cwd = os.getcwd()
        self._initialize_testdir()
        
        try:
            os.chdir(self.TEST_DIR)
            yield
        except:
            traceback.print_exc()
            sys.exit(1)
        finally:
            os.chdir(cwd)
    
    def _get_pytest_config(self, argv):
        """
        Return the ``pytest`` configuration object based on the 
        command-line arguments
        """
        import _pytest.config as _config
        
        # disable pytest-cov
        argv += ['-p', 'no:pytest_cov']
        
        # get the pytest configuration object
        try:
            config = _config._prepareconfig(argv, self.plugins)
        except _config.ConftestImportFailure as e:
            tw = _config.py.io.TerminalWriter(sys.stderr)
            for line in traceback.format_exception(*e.excinfo):
                tw.line(line.rstrip(), red=True)
            tw.line("ERROR: could not load %s\n" % (e.path), red=True)
            raise 
            
        return config
    
    def _initialize_testdir(self):
        """
        Initialize the ``build/test/`` directory
        """
        try:
            shutil.rmtree(self.TEST_DIR)
        except OSError:
            pass
        try:
            os.makedirs(self.TEST_DIR)
        except OSError:
            pass
    
    
    def _fix_test_paths(self, site_dir, args):
        """
        Fix the paths of tests to run to point to the corresponding
        tests in the site directory
        """
        def fix_test_path(x):
            p = x.split('::')
            p[0] = os.path.relpath(os.path.abspath(p[0]), self.ROOT_DIR)
            p[0] = os.path.join(site_dir, p[0])
            return '::'.join(p)
        return [fix_test_path(x) for x in args]
    
    
    def _build_project(self, args):
        """
        Build a dev version of the project.

        Returns
        -------
        site_dir
            site-packages directory where it was installed
        """

        root_ok = [os.path.exists(os.path.join(self.ROOT_DIR, fn))
                   for fn in self.PROJECT_ROOT_FILES]
        if not all(root_ok):
            print("To build the project, run runtests.py in "
                  "git checkout or unpacked source")
            sys.exit(1)

        dst_dir = os.path.join(self.ROOT_DIR, 'build', 'testenv')

        env = dict(os.environ)
        cmd = [sys.executable, 'setup.py']

        # Always use ccache, if installed
        env['PATH'] = os.pathsep.join(self.EXTRA_PATH + env.get('PATH', '').split(os.pathsep))
        cmd += ['build']
        cmd += ['install', '--prefix=' + dst_dir]

        if args.show_build_log:
            ret = subprocess.call(cmd, env=env, cwd=self.ROOT_DIR)
        else:
            log_filename = os.path.join(self.ROOT_DIR, 'build.log')
            print("Building, see build.log...")
            with open(log_filename, 'w') as log:
                p = subprocess.Popen(cmd, env=env, stdout=log, stderr=log,
                                     cwd=self.ROOT_DIR)

        # Wait for it to finish, and print something to indicate the
        # process is alive, but only if the log file has grown (to
        # allow continuous integration environments kill a hanging
        # process accurately if it produces no output)
        last_blip = time.time()
        last_log_size = os.stat(log_filename).st_size
        while p.poll() is None:
            time.sleep(0.5)
            if time.time() - last_blip > 60:
                log_size = os.stat(log_filename).st_size
                if log_size > last_log_size:
                    print("    ... build in progress")
                    last_blip = time.time()
                    last_log_size = log_size

            ret = p.wait()

        if ret == 0:
            print("Build OK")
        else:
            if not args.show_build_log:
                with open(log_filename, 'r') as f:
                    print(f.read())
                print("Build failed!")
            sys.exit(1)

        from distutils.sysconfig import get_python_lib
        site_dir = get_python_lib(prefix=dst_dir, plat_specific=True)
        if not os.path.exists(os.path.join(site_dir, self.PROJECT_MODULE)):
            # purelib?
            site_dir = get_python_lib(prefix=dst_dir, plat_specific=False)
        if not os.path.exists(os.path.join(site_dir, self.PROJECT_MODULE)):
            print("Package %s not properly installed" % self.PROJECT_MODULE)
            sys.exit(1)
        return site_dir
